import logging
from typing import TYPE_CHECKING, List, Optional, Union

import aiohttp

from .lock import DoorLock
from .sensor import LeakSensor
from .switch import BinarySwitch, MultilevelSwitch
from .thermostat import Thermostat
from .utils import Client

_LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    ALL_DEVICE_TYPES = Union[
        DoorLock, Thermostat, BinarySwitch, MultilevelSwitch, LeakSensor
    ]


class API:
    """
    Represents overall SmartRent api

    ``email`` is the email address for your SmartRent account

    ``password`` you know what it is

    ``aiohttp_session`` (optional) uses the aiohttp_session that is passed in

    ``tfa_token`` (optional) tfa token to pass in for login
    """

    def __init__(
        self,
        email: str,
        password: str,
        aiohttp_session: aiohttp.ClientSession = None,
        tfa_token=None,
    ):
        self._device_list: List["ALL_DEVICE_TYPES"] = []
        self.client: Client = Client(email, password, aiohttp_session, tfa_token)

    async def async_fetch_devices(self):
        """
        Fetches list of devices by calling SmartRent api
        """
        _LOGGER.info("Fetching devices via API...")
        data = await self.client.async_get_devices_data()
        _LOGGER.info("Got devices!")

        for device in data:
            device_id = device.get("id")
            device_type = device.get("type")

            device_object: Optional["ALL_DEVICE_TYPES"] = None

            if device_type == "thermostat":
                device_object = Thermostat(device_id, self.client)

            elif device_type == "entry_control":
                device_object = DoorLock(device_id, self.client)

            elif device_type == "switch_binary":
                device_object = BinarySwitch(device_id, self.client)

            elif device_type == "switch_multilevel":
                device_object = MultilevelSwitch(device_id, self.client)

            elif device_type == "sensor_notification":
                attr_names = [attr.get("name") for attr in device.get("attributes")]

                if "leak" in attr_names:
                    device_object = LeakSensor(device_id, self.client)

            if device_object:
                # pass in intial device config
                await device_object._async_fetch_state()

                # add device to device_list
                self._device_list.append(device_object)

    def get_device_list(
        self,
    ) -> List["ALL_DEVICE_TYPES"]:
        """
        Gets list of all devices found
        """
        return self._device_list

    def get_locks(self) -> List[DoorLock]:
        """
        Gets list of DoorLocks
        """
        return self._list_maker(DoorLock)  # type: ignore

    def get_thermostats(self) -> List[Thermostat]:
        """
        Gets list of Thermostats
        """
        return self._list_maker(Thermostat)  # type: ignore

    def get_switches(self) -> List[BinarySwitch]:
        """
        Gets list of BinarySwitches,
        Deprecating soon in favor of ``get_binary_switches``
        """
        _LOGGER.warning(
            'Function "get_switches" will be removed in a future update!'
            ' Please use "get_binary_switches" or "get_multilevel_switches".'
        )
        return self._list_maker(BinarySwitch)  # type: ignore

    def get_binary_switches(self) -> List[BinarySwitch]:
        """
        Gets list of BinarySwitches
        """
        return self._list_maker(BinarySwitch)  # type: ignore

    def get_multilevel_switches(self) -> List[MultilevelSwitch]:
        """
        Gets list of MultilevelSwitches
        """
        return self._list_maker(MultilevelSwitch)  # type: ignore

    def get_leak_sensors(self) -> List[LeakSensor]:
        """
        Gets list of LeakSensors
        """
        return self._list_maker(LeakSensor)  # type: ignore

    def _list_maker(
        self,
        to_find: "ALL_DEVICE_TYPES",
    ) -> List[Optional["ALL_DEVICE_TYPES"]]:
        """
        Gets list of X type of item from ``device_list``
        """
        return [x for x in self._device_list if isinstance(x, to_find)]  # type: ignore


async def async_login(
    email: str,
    password: str,
    aiohttp_session: aiohttp.ClientSession = None,
    tfa_token: str = None,
) -> API:
    """
    Logs into SmartRent and retruns an ``API`` object.

    Prepopulates ``API`` object with devices.

    ``email`` is the email address for your SmartRent account

    ``password`` you know what it is

    ``aiohttp_session`` (optional) uses the aiohttp_session that is passed in

    ``tfa_token`` (optional) tfa token to pass in for login
    """

    smart_rent_api = API(email, password, aiohttp_session, tfa_token)
    await smart_rent_api.async_fetch_devices()

    return smart_rent_api



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smartrent']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.4,<4.0.0', 'websockets>=11.0.1,<12.0.0']

setup_kwargs = {
    'name': 'smartrent-py',
    'version': '0.4.4',
    'description': 'Unofficial Python API for SmartRent devices',
    'long_description': "# SmartRent API\n\n[![PyPI version][pypi-version-badge]](https://pypi.org/project/smartrent-py/)\n[![Supported Python versions][supported-versions-badge]](https://pypi.org/project/smartrent-py/)\n[![PyPI downloads monthly][m-downloads-badge]](https://pypistats.org/packages/smartrent-py)\n[![GitHub License][license-badge]](LICENSE.txt)\n[![Documentation Status][docs-badge]](https://smartrentpy.readthedocs.io/en/latest/?badge=latest)\n[![Code style: black][black-badge]](https://github.com/psf/black)\n\n`smartrent-py` is a simple library for SmartRent devices\n\nUses websockets for communication and supports async functions\n\nFeel free to ⭐️ this repo to get notified about the latest features!\n\n[📚 Read the docs! 📚](https://smartrentpy.readthedocs.io)\n## Supported Devices\nThis client supports:\n* 🔐 Door Locks\n* 🌡 Thermostats\n* 💧 Leak Sensors\n* 💡 Binary Switches\n* 🎚 Multilevel (Dimmer) Switches\n\n\n## Usage\n\n### Installation\n\n```bash\npip install smartrent-py\n```\n\n### Getting an API Object\nIn order to get an api object to interact with, you must login with the `async_login` function. This starts and handles a web session with SmartRent.\n\n```python\nimport asyncio\n\nfrom smartrent import async_login\n\nasync def main():\n    api = await async_login('<EMAIL>', '<PASSWORD>')\n\nasyncio.run(main())\n```\n\n### Getting Devices\nYou can get lists of your devices from the api with the `get_locks`, `get_thermostats`, `get_switches` and `get_leak_sensors` functions. You can then interact with the devices with their getter and setter functions.\n\n```python\nimport asyncio\n\nfrom smartrent import async_login\n\nasync def main():\n    api = await async_login('<EMAIL>', '<PASSWORD>')\n\n    lock = api.get_locks()[0]\n    locked = lock.get_locked()\n\n    if not locked:\n        await lock.async_set_locked(True)\n\nasyncio.run(main())\n```\n\n### Automatic Updating\nIf you need to get live updates to your device object from SmartRent, you can do that by calling `start_updater`. You can stop getting updates by calling `stop_updater`.\n\nYou can also set a callback function via `set_update_callback` that will be called when an update is triggered.\n\nFor example, if you want to set your thermostat to `Dad Mode` you can trigger an event every time the `cooling_setpoint` is changed and just change it back to your own desired value.\n```python\nimport asyncio\n\nfrom smartrent import async_login\n\nasync def main():\n    api = await async_login('<EMAIL>', '<PASSWORD>')\n\n    thermo = api.get_thermostats()[0]\n    thermo.start_updater()\n\n    CONSTANT_COOL = 80\n\n    async def on_evt():\n        if CONSTANT_COOL != thermo.get_cooling_setpoint():\n            await thermo.async_set_cooling_setpoint(CONSTANT_COOL)\n\n    thermo.set_update_callback(on_evt)\n\n    while True:\n        await asyncio.sleep(60)\n\nasyncio.run(main())\n```\n\n## Development\n### Setting up dev enviornment\n\n```\npip install -r requirements_test.txt\n```\n\n### Running the code formatter\n[Black](https://github.com/psf/black) is used for quick and easy code formatting\n\n```\nblack smartrent\n```\n\n## Special thanks\nMuch inspiration was taken from these projects:\n\n[AMcPherran/SmartRent-MQTT-Bridge](https://github.com/AMcPherran/SmartRent-MQTT-Bridge)\n\n[natanweinberger/smartrent-python](https://github.com/natanweinberger/smartrent-python)\n\n[Burry/homebridge-smartrent](https://github.com/Burry/homebridge-smartrent)\n## Versioning\n\nWe use [SemVer](http://semver.org/) for versioning. For the versions available, see the [tags on this repository](https://github.com/your/project/tags).\n\n# Adding Unsupported Devices!\n\nThis library is built in a way that it is easy to add any smartrent device. If you find that one of the devices you own is not supported, we should be able to add it to this libray.\n\nTwo things are needed to get the info to add a new device:\n* We have to get some device info from smartrents api\n* We need to get info about how the device reacts when it enters different states\n\nBoth of these can be done in one script.\n\nYou have to run `python3 ./tools/device_helper.py`\n\nIt will print out some output and a table with `Device ID`s and `Device Names`. The device you wish to add should be listed in that table.\n```\nDevice ID:      Device Name:\n==============================\n2211234         Thermostat\n2212345         Front Door - Lock\n2212347         Washer - Leak Sensor\n```\n\nThe program will then ask for a `Device ID` that you wish to track. Enter in the `Device ID` you want to gather information from.\n\n```\nPut in the Device ID you wish to track: 2211234\nJoining topic for 2211234...\n{'response': {}, 'status': 'ok'}\n```\n\nYou can now change the state of your device either physically in your apartment/house or through the SmartRent app. For example, this is a thermostat's cooling point being turned up, and then the thermostat sending an update about the current humidity\n```\nThermostatSetpoint  cooling_setpoint    77\nMultiLvlSensor      current_humidity    53\n```\n\nOnce you get a good amount of info, open up an issue describing that you wish to add support for a new device and paste all of the output from `device_helper.py`. Feel free to hide your `Device ID`s if you wish.\n\n\n[pypi-version-badge]: https://img.shields.io/pypi/v/smartrent-py.svg?logo=pypi&logoColor=FFE873&style=for-the-badge\n[supported-versions-badge]: https://img.shields.io/pypi/pyversions/smartrent-py.svg?logo=python&logoColor=FFE873&style=for-the-badge\n[downloads-badge]: https://static.pepy.tech/personalized-badge/smartrent-py?period=total&units=international_system&left_color=grey&right_color=orange&left_text=total%20downloads&style=for-the-badge\n[m-downloads-badge]: https://img.shields.io/pypi/dm/smartrent-py.svg?style=for-the-badge\n[license-badge]: https://img.shields.io/github/license/ZacheryThomas/smartrent-py.svg?style=for-the-badge\n[docs-badge]: https://readthedocs.org/projects/smartrentpy/badge/?version=latest&style=for-the-badge\n[black-badge]: https://img.shields.io/badge/code%20style-black-000000.svg?style=for-the-badge\n",
    'author': 'Zachery Thomas',
    'author_email': 'zacherythomas12@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zacherythomas/smartrent-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import inspect
import logging
from collections.abc import Callable
from typing import Any, Dict, List, Optional, Union

from .utils import Client

_LOGGER = logging.getLogger(__name__)


class Device:
    """
    Base class for SmartRent devices
    """

    def __init__(self, device_id: Union[str, int], client: Client):
        self._device_id = int(device_id)
        self._name: str = ""
        self._online: Optional[bool] = None
        self._battery_powered: Optional[bool] = None
        self._battery_level: Optional[int] = None
        self._update_callback_funcs: List[Callable[[None], None]] = []

        self._client: Client = client

    def __del__(self):
        self.stop_updater()

    def get_name(self) -> Optional[str]:
        """
        Gets the name of the device, if known
        """
        return self._name

    def get_online(self) -> Optional[bool]:
        """
        Gets if device is online or not
        """
        return self._online

    def get_battery_powered(self) -> Optional[bool]:
        """
        Gets if devices is battery powered
        """
        return self._battery_powered

    def get_battery_level(self) -> Optional[int]:
        """
        Gets devices battery level (assuming device is battery powered)
        """
        return self._battery_level

    @staticmethod
    def _structure_attrs(attrs: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Converts device json object to hirearchical list of attributes

        ``attrs``: List of device attributes
        """
        structure: Dict[str, Any] = {}

        for attr in attrs:
            name: str = str(attr.get("name"))
            state: str = str(attr.get("state"))

            structure[name] = state

        _LOGGER.info("constructed attribute structure: %s", structure)
        return structure

    def _fetch_state_helper(self, data: Dict[str, str]):
        """
        Called by ``_async_fetch_state``

        Converts event dict to device param info
        """
        raise NotImplementedError

    async def _async_fetch_state(self):
        """
        Fetches device information from SmartRent api

        Calls ``_fetch_state_helper`` so device can parse out
        info and update it's state.

        Calls function passed into ``set_update_callback`` if it exists.
        """
        _LOGGER.info("%s: Fetching data from device endpoint...", self._name)
        device_data = await self._client.async_get_device_data(self._device_id)

        device_at_start = dict(vars(self))

        self._battery_level = device_data.get("battery_level")
        self._battery_powered = device_data.get("battery_powered")
        self._online = device_data.get("online")

        self._fetch_state_helper(device_data)

        device_at_end = dict(vars(self))

        # If device attrs updated, call callbacks
        if not device_at_start == device_at_end:
            await self._async_call_callbacks()

    def start_updater(self):
        """
        Allows device to update it's attrs in the background
        """
        self._client._subscribe_device_to_updater(self)

    def stop_updater(self):
        """
        Turns off automatic attr updates
        """
        self._client._unsubscribe_device_to_updater(self)

    def set_update_callback(self, func) -> None:
        """
        Allows callback to be fired when ``Client._async_update_state``
        or ``_async_fetch_state`` gets new information
        """

        self._update_callback_funcs.append(func)

    def unset_update_callback(self, func) -> None:
        """
        Removes callback from being fired when ``Client._async_update_state``
        or ``_async_fetch_state`` gets new information
        """
        try:
            self._update_callback_funcs.remove(func)
        except ValueError:
            pass

    def _update_parser(self, event: Dict[str, Any]) -> None:
        """
        Called by ``Client._async_update_state``

        Converts event dict to device attr info
        """
        raise NotImplementedError

    async def _update(self, event: Dict[str, Any]):
        """
        Recieves event dict, calls ``_update_parser`` for each device and callbacks
        """
        # handle updating of device attrs
        self._update_parser(event)

        # handle calling callbacks
        await self._async_call_callbacks()

    async def _async_call_callbacks(self):
        """
        Handles calling all callbacks
        """
        for func in self._update_callback_funcs:
            if inspect.iscoroutinefunction(func):
                await func()
            else:
                func()
